"""To import in your own project:
#$ from textual-pyfiglet import FigletWidget

To install the extended fonts collection:
#$ pip install textual-pyfiglet[fonts]

I made sure to preserve the original PyFiglet CLI for reference.
You can access the original CLI with the following command:
#$ python -m textual_pyfiglet.pyfiglet

The original PyFiglet CLI has a demo that can be accessed like this (verbatim, with the spaces):
#$ python -m textual_pyfiglet.pyfiglet some text here

In order to change fonts with the original demo, you can use the -f flag like this:
#$ python -m textual_pyfiglet.pyfiglet -f slant Hey hows it going?

To see all installed fonts (scans fonts folder):
#$ python -m textual_pyfiglet.pyfiglet --list_fonts

# NOTE: You can also download FIG fonts from the internet and just drop them in the fonts folder.
See the readme for more information.
"""

import os
import shutil
import importlib
import configparser

from .figletwidget import FigletWidget
from .pyfiglet import fonts

config = configparser.ConfigParser()
try:
    with importlib.resources.path(__package__, 'config.ini') as config_path:
        cfgfile = config.read(config_path)
except FileNotFoundError:
    raise FileNotFoundError("config.ini not found")
if not cfgfile:
    raise FileNotFoundError("config.ini not found")

extended_fonts_installed = config.getboolean('DEFAULT', 'extended_fonts_installed')

def check_for_extended_fonts():
    """If the extended fonts are not installed, this function will check if they've been installed.
    This check will run every time the package is imported, so you can install the fonts at any time."""

    package_name = "textual_pyfiglet_fonts"

    # check if the fonts package is installed
    if importlib.util.find_spec(package_name) is not None:
        fonts_module = importlib.import_module(package_name)
    else:
        return

    xtra_fonts_path = os.path.dirname(fonts_module.__file__)    # get the path to the xtra fonts folder
    fonts_folder = os.path.dirname(fonts.__file__)              # get the path to the main fonts folder

    # get list of files in xtra_fonts_path
    xtra_fonts = os.listdir(xtra_fonts_path)

    # copy all fonts to the fonts folder
    for font in xtra_fonts:
        font_path = os.path.join(xtra_fonts_path, font)
        new_font_path = os.path.join(fonts_folder, font)
        try:
            shutil.copyfile(font_path, new_font_path)
        except Exception as e:
            print(f"Error copying font: {font} - {e}")

    config['DEFAULT']['extended_fonts_installed'] = "True"

    # Write the changes back to the INI file
    with importlib.resources.path(__package__, 'config.ini') as config_file:
        with open(config_file, 'w') as f:
            config.write(f)


if not extended_fonts_installed:
    check_for_extended_fonts()










from __future__ import annotations

from .pyfiglet import Figlet

from textual.app import App, on
from textual.message import Message
from textual.containers import Horizontal, Container, VerticalScroll
from textual.widgets import Header, Footer, Button, Static, TextArea, Select


class FigletWidget(Static):
    """Adds simple PyFiglet ability to the Static widget.    
    NOTE: The width and height are set to auto, so the widget will expand to fit the text.    
    The size of the PyFiglet widget can vary greatly depending on the font and text.    
    The default is auto so it will expand or contract automatically.    
    In your own app you might wish to change this to a fixed size.    
    See __init__ for more details."""

    DEFAULT_CSS = """
    FigletWidget {
        width: auto;
        height: auto;
    }
    """

    class Updated(Message):
        """This is here to provide a message to the app that the widget has been updated.
        You might need this to trigger something else in your app resizing, adjusting, etc.
        The size of FIG fonts can vary greatly, so this might help you adjust other widgets."""

        def __init__(self, widget: FigletWidget) -> None:
            super().__init__()
            self.widget = widget

        @property
        def control(self) -> FigletWidget:
            """This is required to be able to use the 'selector' property
            when using the message handler."""

            return self.widget



    def __init__(self, *args, font: str = "calvin_s", **kwargs) -> None:
        """A custom widget for turning text into ASCII art using PyFiglet.
        This args section is copied from the Static widget. It's the same except for the font argument.

        This class is designed to be an easy drop in replacement for the Static widget.
        The only new argument is 'font', which has a default set to one of the smallest fonts.
        You can replace any Static widget with this and it should just work (aside from the size)

        Args:
            renderable: A Rich renderable, or string containing console markup.
            font (PyFiglet): Font to use for the ASCII art. Default is "calvin_s".
            expand: Expand content if required to fill container.
            shrink: Shrink content if required to fill container.
            markup: True if markup should be parsed and rendered.
            name: Name of widget.
            id: ID of Widget.
            classes: Space separated list of class names.
            disabled: Whether the static is disabled or not.

        Included fonts:
        - standard
        - small
        - calvin_s
        - slant
        - small slant
        - modular
        - chunky
        - broadway_kb
        - cybermedium

        Remember you can always download more fonts. To download the extended fonts pack:
        # TODO Add me here when I have the link

        You can also download individual fonts online and drop them in the fonts folder.
        """
        super().__init__(*args, **kwargs)
        self.stored_text = str(self.renderable)
        self.font = font
        self.screen_width = self.app.size.width
        self.figlet = Figlet(font=font, width=self.screen_width - 8)

        # NOTE: Figlet also has "direction" and "justify" arguments,
        # but I'm not using them here yet. Should probably be added in the future.
        # TODO Add Direction and Justify arguments

    def on_mount(self):
        self.update()

    def update(self, new_text: str | None = None):
        """Update the PyFiglet area with the new text.    
        Note that this over-rides the standard update method in the Static widget!   
        This does NOT take any rich renderable like the Static widget does.
        It can only take a text string.

        Args:
            new_text: The text to update the PyFiglet widget with. Default is None.
        """

        if new_text is not None:
            self.stored_text = new_text
        self.renderable = self.figlet.renderText(self.stored_text)

        # this line is important
        # this makes textual reset the widget size to whatever the new renderable is
        self.set_styles("width: auto; height: auto;")   
        
        # Necessay to refresh the widget to apply the changes
        self.refresh()

        # Post a message to the app that the widget has been updated
        self.post_message(self.Updated(self))

    def change_font(self, font: str) -> None:
        self.figlet.setFont(font=font)
        self.update()

    def change_width(self, width: int) -> None:
        self.figlet.width = width
        self.update()



"""Contains the demo app.
This module contains the demo application for PyFiglet using the Textual framework.
Classes:
    PyFigletDemo(App): A Textual application demonstrating the usage of PyFiglet with various fonts and text input.
Functions:
    get_all_fonts_list(package_path: str) -> list: Returns a list of all font filenames (without extensions) in the specified package path.
Usage:
    Run this module directly to start the PyFiglet demo application.
    Example: python -m textual-pyfiglet
Notes:
    - The application uses Rich for enhanced traceback formatting.
    - The application defines a default CSS for styling the layout and widgets.
    - The font list is dynamically generated from the specified directory.
    - The application includes a timer to set initial text in the TextArea to avoid glitches.
    - The application handles font changes and text input changes to update the FigletWidget accordingly."""

import os


from rich import traceback
traceback.install()


from textual.app import App, on
from textual.containers import Horizontal, Container, VerticalScroll
from textual.widgets import Header, Footer, Button, Static, TextArea, Select, Switch, Label

# NOTE: There are HUNDREDS of fonts available for Figlet.
# The fonts included are just a small selection, to keep the library size down.
# Also note, you can simply download more and add them.
# Any FIGlet fonts you drop in the fonts folder will be automatically loaded by PyFiglet
# as well as this demo. The selection list in the demo will update automatically.


from .figletwidget import FigletWidget
from .pyfiglet import fonts


def get_all_fonts_list() -> list:
    """Scans the fonts folder.
    Returns a list of all font filenames (without extensions)."""

    # first get the path of the fonts package:
    package_path = os.path.dirname(fonts.__file__)
    path_list = os.listdir(package_path)
    fonts_list = []
    for filename in path_list:
        if not filename.endswith('.py') and not filename.startswith('__pycache__'):
            fonts_list.append(os.path.splitext(filename)[0])
    return fonts_list


class PyFigletDemo(App):


    DEFAULT_CSS = """
    #main_content {
        align: center middle;
        width: 1fr;
        height: 1fr;
    }

    #font_select {
        width: 22;
    }

    #options_bar {
        content-align: center middle;
        align: center middle;
        height: 4;
        padding: 0;
        background: $boost;
    }

    TextArea {
        height: auto
    }

    FigletWidget {
        background: $boost;
        padding: 1 4 0 4;
        content-align: center middle;
    }
    """

    fonts_list = get_all_fonts_list()
    fonts_list.sort()
    font_options =  [(font, font) for font in fonts_list]

    def compose(self):

        self.log.info("PyFiglet Demo started.")
        self.log.debug(f"Available fonts: {self.fonts_list}")

        yield Header("PyFiglet Demo")

        with VerticalScroll(id="main_content"):
            yield FigletWidget("Hello, World!", id="figlet", font="standard")
        with Horizontal(id="options_bar"):
            yield Select(options=self.font_options, value="standard", id="font_select", allow_blank=False)
            yield TextArea("", id="text_input")

        yield Footer()

    def on_mount(self):

        self.set_timer(0.05, self.set_starter_text)
        # The timer is because starting with text in the TextArea makes it glitch out.
        # Giving it a 50ms delay to set the text fixes the problem.

    # This just sets the cursor to the end of the text in the TextArea when the app starts:
    def set_starter_text(self):
        self.query_one("#text_input").text = "Hello, World!"
        self.query_one("#text_input").focus()
        end = self.query_one("#text_input").get_cursor_line_end_location()
        self.query_one("#text_input").move_cursor(end)
        # self.query_one("#options_bar").set_styles("height: auto;")

    def on_resize(self, event):
        width, height = event.size      # TODO This needs to be tested in different terminals and app environments
        self.query_one("#figlet").change_width(width-8)    # -8 to account for padding

    @on(Select.Changed, selector="#font_select")           
    def font_changed(self, event: Select.Changed) -> None:
        self.query_one("#figlet").change_font(event.value)

    @on(TextArea.Changed)
    async def text_changed(self):
        text = self.query_one("#text_input").text
        self.query_one("#figlet").update(text)  # Thats it! The FigletWidget will update the text for you.

        # This just scrolls the text area to the end when the text changes:
        scroll_area = self.query_one("#main_content")
        if scroll_area.scrollbars_enabled == (True, False):
            scroll_area.action_scroll_end()

# This is for the entry point script. You can run the demo with:
#$ textual-pyfiglet
def main():
    app = PyFigletDemo()
    app.run()
 
# TODO remove me when confirmed working everywhere
# This is another way to run the demo
#$ python -m textual-pyfiglet 
if __name__ == "__main__":
    app = PyFigletDemo()
    app.run()